"""Generate shareable Explorer URL"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import GetInstanceURL


@click.command("url")
@click.option(
    "--expiration-seconds",
    default=(24 * 60 * 60),  # 24 hours
    help="Number of seconds before the URL expires. Maximum is 24h (86400s).",
    show_default=True,
)
@add_options(common_options)
def url(  # pylint: disable=too-many-arguments
    expiration_seconds: int,
    host: str,
    email: str,
    password: str,
    api_key: str,
):  # pylint: disable=line-too-long
    """Generate shareable URL for Explorer instance."""  # noqa: E501
    GetInstanceURL(
        expiration_seconds,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Utilities for processing copied samples."""


def get_line(copied_sample):
    """Build a list of relevant data to be printed.

    Args:
        copied_sample (SampleCopy): instance of SampleCopy

    Returns:
        str: relevant data to be printed
    """

    return "\t".join(
        [
            str(copied_sample.sample_id),
            copied_sample.client_id if copied_sample.client_id else "",
        ]
    )



"""Configure inactivity stop for explorer instances subcommand."""

from ....base import Command
from .....exceptions import ValidationError


class GetInstanceURL(Command):
    """Get explorer instance URL command executor."""

    def __init__(self, expiration_seconds, credentials, options):
        super().__init__(credentials, options)
        self.expiration_seconds = expiration_seconds

    def validate(self):
        """Validate get URL"""

    def initialize(self):
        """Initialize get URL subcommand."""
        self.login()

    def execute(self):
        """Make a request to get explorer instance URL."""
        self.echo_debug("Get explorer instance URL.")

        explorer_instances = self.api_client.get_explorer_instances()
        self.echo_debug(f"Found {len(explorer_instances.results)} explorer instances.")

        if len(explorer_instances.results) != 1:
            raise ValidationError(
                "Command not supported. Download the latest version of the Gencove CLI."
            )

        if explorer_instances.results[0].status != "running":
            raise ValidationError(
                "Instance is not running. Cannot retrieve access URL."
            )

        instance_id = str(explorer_instances.results[0].id)
        self.echo_info("Request to generate explorer access URL accepted.")
        url = self.api_client.get_explorer_access_url(
            instance_id=instance_id,
            access_token_expiration=self.expiration_seconds,
        )
        self.echo_info(
            f"Explorer access URL "
            f"(valid for {self.expiration_seconds} seconds):\n{url.url}"
        )



"""Utilities for different levels of click.echo outputs."""
import datetime
import logging
import os
import sys
import uuid
from platform import platform

import boto3

import click

from gencove.version import version

try:
    utc_tz = datetime.UTC  # Python 3.11+ only
except AttributeError:
    utc_tz = datetime.timezone.utc  # fallback for older Python versions


INFO = "INFO"
DEBUG = "DEBUG"
LOG_LEVEL = os.environ.get("GENCOVE_LOG", INFO)
DEBUG_LOG = []


def _echo(msg, **kwargs):
    """Output click echo message."""
    if LOG_LEVEL == DEBUG:
        msg = f"{datetime.datetime.now(utc_tz).isoformat()} {msg}"
    click.echo(msg, **kwargs)


def _log(msg):
    """Adds the message to the log list."""
    msg = f"{datetime.datetime.now(utc_tz).isoformat()} {msg}"
    DEBUG_LOG.append(msg)


def output_warning(text):
    """Click echo warning."""
    return click.style(f"WARNING: {text}", fg="bright_yellow", bold=True)


def output_error(text):
    """Click echo error."""
    return click.style(f"ERROR: {text}", fg="bright_red", bold=True)


def echo_data(msg, **kwargs):
    """Output click echo msg."""
    _echo(msg, **kwargs)
    _log(msg)


def echo_info(msg, **kwargs):
    """Output click echo msg."""
    _echo(msg, err=True, **kwargs)
    _log(msg)


def echo_debug(msg, **kwargs):
    """Output click echo msg only if debug is on."""
    if LOG_LEVEL == DEBUG:
        _echo(msg, err=True, **kwargs)
    _log(msg)


def echo_warning(msg, **kwargs):
    """Output click echo msg with background."""
    _echo(output_warning(msg), err=True, **kwargs)
    _log(msg)


def echo_error(msg, **kwargs):
    """Output click echo msg with background."""
    _echo(output_error(msg), err=True, **kwargs)
    _log(msg)


echo_debug(
    f"Python version: {sys.version_info.major}."
    f"{sys.version_info.minor}."
    f"{sys.version_info.micro}"
)
echo_debug(f"CLI version: {version()}")
echo_debug(f"OS details: {platform()}")
echo_debug(f"boto3 version: {boto3.__version__}")
if LOG_LEVEL == DEBUG:
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    logging.getLogger("botocore.auth").setLevel(logging.CRITICAL)


def get_debug_file_name():
    """Creates a 'unique' file name YYYY_MM_DD_HH_MM_RANDOM_ID.log
    inside a YYYY_MM folder.
    """
    now = datetime.datetime.now(utc_tz)
    timestamp = f"{now:%Y_%m_%d_%H_%M_%S}"
    random_id = str(uuid.uuid4())[:8]
    filename = f"{timestamp}_{random_id}.log"
    folder_name = f"{now:%Y_%m}"
    folder = os.path.join(".logs", folder_name)
    file_path = os.path.join(folder, filename)
    if not os.path.exists(folder):
        os.makedirs(folder)
    return file_path


def dump_debug_log():
    """Saves the logs into a file."""
    if os.environ.get("GENCOVE_SAVE_DUMP_LOG") != "FALSE":
        try:
            debug_filename = get_debug_file_name()
            log = "\n".join(DEBUG_LOG)
            with open(debug_filename, "w", encoding="utf-8") as dump_file:
                dump_file.write(log)
            _echo(
                f"Please attach the debug log file located in {debug_filename} to a bug report."  # noqa: E501  # pylint: disable=line-too-long
            )
        except PermissionError:
            echo_warning(
                "An error occurred, but we couldn't write a debug log file to disk."  # noqa: E501  # pylint: disable=line-too-long
            )



"""Utility to generate version string."""
import os


def version():
    """Return version string."""
    base_dir = os.path.dirname(__file__)
    with open(
        os.path.join(base_dir, "version", "A-major"), encoding="utf-8"
    ) as major_file:
        major = major_file.read().strip()
    with open(
        os.path.join(base_dir, "version", "B-minor"), encoding="utf-8"
    ) as minor_file:
        minor = minor_file.read().strip()
    with open(
        os.path.join(base_dir, "version", "C-patch"), encoding="utf-8"
    ) as patch_file:
        patch = patch_file.read().strip()
    return f"{major}.{minor}.{patch}"



"""
collections_extras

Provides collection utilities, including custom lazy evaluation mechanisms.
"""

from typing import Generator, Generic, List, TypeVar


T = TypeVar("T")


class LazyList(Generic[T]):
    """
    A lazily-evaluated, cache-backed list-like container.

    LazyList allows the use of a generator to lazily compute elements on demand.
    Items are cached as they are accessed, and the list can be iterated multiple
    times without re-generating values.

    Attributes:
        _generator (Generator[T, None, None]): The underlying generator providing
            elements.
        _cache (List[T]): Cached elements for previously computed indices.
        _exhausted (bool): Tracks whether the generator has been completely iterated.

    Methods:
        __getitem__(index): Accesses elements by index, computing and caching as
            needed.
        __iter__(): Iterates over cached and computed elements.
        __len__(): Exhausts the generator and returns the total number of computed
            elements.
        __repr__(): Provides a string representation of the LazyList.
    """

    def __init__(self, generator: Generator[T, None, None]):
        self._generator: Generator[T, None, None] = generator
        self._cache: List[T] = []
        self._exhausted: bool = False

    def __getitem__(self, index: int) -> T:
        while not self._exhausted and len(self._cache) <= index:
            try:
                self._cache.append(next(self._generator))
            except StopIteration:
                self._exhausted = True

        if index < len(self._cache):
            return self._cache[index]
        raise IndexError("LazyList index out of range")

    def __iter__(self):
        for value in self._cache:
            yield value

        while not self._exhausted:
            try:
                value = next(self._generator)
                self._cache.append(value)
                yield value
            except StopIteration:
                self._exhausted = True

    def __len__(self) -> int:
        if not self._exhausted:
            self._cache.extend(self._generator)
            self._exhausted = True
        return len(self._cache)

    def __repr__(self) -> str:
        if self._exhausted:
            return f"LazyList({self._cache})"
        return f"LazyList({self._cache} + [...])" if self._cache else "LazyList([...])"



"""Download project's jointcalled VCF files executor."""
import os

import backoff

from ... import download
from ...base import Command
from .... import client


class GetJointcalledVCF(Command):
    """Download project's jointcalled VCF files executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        project_id,
        output_folder,
        credentials,
        options,
        no_progress,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.output_folder = output_folder or "."
        self.no_progress = no_progress

    def initialize(self):
        """Initialize subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=5,
        max_time=30,
    )
    def execute(self):
        """Download jointcalled VCF files for a given project."""
        self.echo_debug(
            f"Downloading jointcalled VCF files for project {self.project_id}"
        )
        try:
            project = self.api_client.get_project(self.project_id)
            self.echo_debug(project)
            if not project.files:
                self.echo_error(
                    f"Project {self.project_id} has no jointcalled VCF files."
                )
                return

            for file in project.files:
                if file.file_type in ["jointcalled-vcf", "jointcalled-vcf_indexed"]:
                    self.echo_debug(
                        f"Downloading jointcalled VCF file {file.file_type}"
                    )
                    filename = download.utils.get_filename_from_download_url(
                        file.download_url
                    )
                    download_path = os.path.join(self.output_folder, filename)
                    self.echo_debug(
                        f"Downloading jointcalled VCF file to {download_path}"
                    )
                    download.utils.download_file(
                        download_path,
                        file.download_url,
                        no_progress=self.no_progress,
                    )
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 404:
                self.echo_error(f"Project {self.project_id} does not exist.")
            raise



import re

from gencove.tests.filters import _replace_uuid_from_url
from gencove.tests.utils import MOCK_UUID


def filter_project_qc_report_request(request):
    """Filter project UUID data from request."""
    return _replace_uuid_from_url(request, "project-qc-report")


def filter_monthly_usage_report_request(request):
    """Filter project UUID data from request."""
    return _replace_uuid_from_url(request, "monthly-usage")


def filter_report_response_body(response):
    """Filter out all UUIDs in the response body"""
    body = response["body"]["string"].decode()
    uuid_pattern = (
        r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}"
    )
    censored_body = re.sub(uuid_pattern, MOCK_UUID, body)
    response["body"]["string"] = censored_body.encode()
    return response


def filter_report_response_filename(response):
    if "Content-Disposition" in response["headers"]:
        response["headers"]["Content-Disposition"] = [
            'attachment; filename="report.csv"'
        ]
    return response



import json

from gencove.tests.filters import _replace_uuid_from_url
from gencove.tests.utils import MOCK_UUID


def filter_get_sample_manifest_response(response):
    if response["status"]["code"] == 404:
        return response
    try:
        body = json.loads(response["body"]["string"])
        body["id"] = MOCK_UUID
        body["file_name"] = "example.csv"
        body["file"] = {
            "id": MOCK_UUID,
            "s3_path": "s3://example/example.csv",
            "size": 100,
            "download_url": "https://example.com/example.csv",
            "file_type": "sample_manifest-csv",
            "checksum_256": "",
        }
        body["project"] = MOCK_UUID
        body["created"] = "2023-08-28T14:50:34.455552Z"
        response["body"]["string"] = json.dumps(body).encode()
    except json.decoder.JSONDecodeError:
        pass
    return response


def filter_sample_manifest_request(request):
    """Filter project UUID data from request."""
    return _replace_uuid_from_url(request, "sample-manifests")



"""Copy existing samples shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid, validate_uuid_list
from gencove.constants import Credentials, Optionals

from .main import CopyExistingSamples


@click.command("copy-existing-samples")
@click.argument("project_id", callback=validate_uuid)
@click.option(
    "--source-project-id",
    "--project-id",
    required=False,
    help=(
        "Copy all available samples, in succeeded or failed_qc state that have"
        " files, from source project."
        " Either --source-project-id or --source-sample-ids option must be provided,"
        " but not both."
    ),
)
@click.option(
    "--source-sample-ids",
    "--sample-ids",
    required=False,
    help=(
        "A comma separated list of source sample ids to copy into the provided"
        " project."
        " Either --source-project-id or --source-sample-ids option must be provided,"
        " but not both."
    ),
    callback=validate_uuid_list,
)
@add_options(common_options)
def copy_existing_project_samples(  # pylint: disable=too-many-arguments
    project_id,
    source_project_id,
    source_sample_ids,
    host,
    email,
    password,
    api_key,
):  # pylint: disable=line-too-long
    """Copy existing samples to a project.

    `PROJECT_ID`: Gencove project ID

    Examples:

        Copy all available samples from source project:

            gencove projects copy-existing-samples d9eaa54b-aaac-4b85-92b0-0b564be6d7db --source-project-id d8eb0bb5-29ee-44ed-b681-0fc05a557183

        Copy samples by id:

            gencove projects copy-existing-samples d9eaa54b-aaac-4b85-92b0-0b564be6d7db --sample-ids 59f5c1fd-cce0-4c4c-90e2-0b6c6c525d71,7edee497-12b5-4a1d-951f-34dc8dce1c1d
    """  # noqa: E501
    CopyExistingSamples(
        project_id,
        source_project_id,
        source_sample_ids,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Project get jointcalled VCF shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid
from gencove.constants import Credentials, Optionals

from .main import GetJointcalledVCF


@click.command("get-jointcalled-vcf")
@click.argument("project_id", callback=validate_uuid)
@click.option(
    "--output-folder",
    help="Output folder for jointcalled VCF files.",
    type=click.Path(file_okay=False, writable=True, dir_okay=True),
    required=False,
    default=None,
)
@add_options(common_options)
@click.option(
    "--no-progress",
    is_flag=True,
    help="If specified, no progress bar is shown.",
)
# pylint: disable=too-many-arguments
def get_jointcalled_vcf(
    project_id,
    output_folder,
    host,
    email,
    password,
    api_key,
    no_progress,
):
    """Download jointcalled VCF files in a project.

    `PROJECT_ID`: Gencove project ID
    """
    GetJointcalledVCF(
        project_id,
        output_folder,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
        no_progress,
    ).run()



"""Tests for collections_extras of Gencove CLI."""
from collections_extras import LazyList


def test_lazy_evaluation():
    """Test that the generator is evaluated lazily."""

    def gen():
        yield 1
        yield 2
        yield 3

    lazy_list = LazyList(gen())
    assert len(lazy_list._cache) == 0  # pylint: disable=protected-access
    assert lazy_list[0] == 1
    assert len(lazy_list._cache) == 1  # pylint: disable=protected-access


def test_indexing():
    """Test indexing access."""

    def gen():
        yield from range(5)

    lazy_list = LazyList(gen())
    assert lazy_list[2] == 2
    assert lazy_list[4] == 4
    try:
        _ = lazy_list[5]
    except IndexError:
        pass
    else:
        assert False, "Expected IndexError not raised"


def test_iteration():
    """Test iteration over the LazyList."""

    def gen():
        yield from range(5)

    lazy_list = LazyList(gen())
    result = list(lazy_list)
    assert result == [0, 1, 2, 3, 4]
    assert len(lazy_list) == 5


def test_caching():
    """Test that values are cached after access."""

    def gen():
        yield from range(3)

    lazy_list = LazyList(gen())
    _ = lazy_list[1]
    assert lazy_list._cache == [0, 1]  # pylint: disable=protected-access


def test_repr():
    """Test the string representation of LazyList."""

    def gen():
        yield 1
        yield 2
        yield 3

    lazy_list = LazyList(gen())
    assert repr(lazy_list) == "LazyList([...])"
    _ = lazy_list[0]
    assert repr(lazy_list) == "LazyList([1] + [...])"
    _ = list(lazy_list)
    assert repr(lazy_list) == "LazyList([1, 2, 3])"


def test_exhaustion():
    """Test behavior after the generator is exhausted."""

    def gen():
        yield from range(3)

    lazy_list = LazyList(gen())
    _ = list(lazy_list)
    assert lazy_list._exhausted  # pylint: disable=protected-access
    try:
        _ = lazy_list[5]
    except IndexError:
        pass
    else:
        assert False, "Expected IndexError not raised"


def test_len():
    """Test the length of LazyList after generator exhaustion."""

    def gen():
        yield from range(4)

    lazy_list = LazyList(gen())
    assert len(lazy_list) == 4


def test_empty_generator():
    """Test LazyList with an empty generator."""

    def gen():
        if False:  # pylint: disable=using-constant-test
            yield

    lazy_list = LazyList(gen())
    assert len(lazy_list) == 0
    try:
        _ = lazy_list[0]
    except IndexError:
        pass
    else:
        assert False, "Expected IndexError not raised"
    assert not list(lazy_list)


def test_large_generator():
    """Test LazyList with a large generator."""

    def gen():
        for i in range(1_000_000):
            yield i + 1

    lazy_list = LazyList(gen())
    assert lazy_list[999] == 1000
    assert len(lazy_list._cache) < 1_000_000  # pylint: disable=protected-access


def test_iter():
    """Test LazyList iterable."""

    def gen_fib():
        fib_a, fib_b = 1, 1
        while fib_a < 10:
            yield fib_a
            fib_a, fib_b = fib_b, fib_b + fib_a

    lazy_list = LazyList(gen_fib())
    for lazy_result, expected_value in zip(lazy_list, [1, 1, 2, 3, 5, 8, 13]):
        assert lazy_result == expected_value
